#include <fstream>
#include <iostream>
#include "vector.h"
struct Vector4 { //four-dimensional vector only for returning numbers from a function
	float x, y, z, t; //three-dimensional vectors are covered thoroughly in vector.h
};

float max(float A, float B) {
	return (A>B)?A:B;
}
float min(float A, float B) {
	return (A<B)?A:B;
}

class nurbs_surface { //each nurbs_surface stores data for and can export one patch surface
private:
	Vector** P;//Points array
	float** W;//Weights array
	float* Tu;//U knots array
	float* Tv;//V knots array
	int usize, vsize, udegree, vdegree;
	int uknots, vknots, ures, vres;
	char* filehandle;
	
public:
	nurbs_surface() {}
	~nurbs_surface() {
		for(int i=0; i<usize; i++) {
			delete [] P[i];
			delete [] W[i];
		}
		delete [] P;
		delete [] W;
		//delete [] Tu; //not sure about memory leaks...
		//delete [] Tv;
	}
/**************************************/
/********ROUTINE TO IMPORT FILE********/
/**************************************/
	void import_data(char* fname) {
		filehandle = new char [64];
		strcpy(filehandle,fname);
		strtok(filehandle,".");
		cout << "file handle = " << filehandle << endl;
		cout << "importing data from " << fname << endl;
		FILE* fptr;
		fptr = fopen(fname,"r");
		char letter1, letter2;
		char fstring [64];
		
		rewind(fptr);
		fscanf(fptr,"%i %i %i %i %i %i",&usize,&vsize,&udegree,&vdegree,&ures,&vres);
		
		uknots = usize+udegree+1;
		vknots = vsize+vdegree+1;
		cout << usize << " u points" << endl;
		cout << vsize << " v points" << endl;
		cout << udegree << " = u degree" << endl;
		cout << vdegree << " = v degree" << endl;
		cout << uknots << " = u knots" << endl;
		cout << vknots << " = v knots" << endl;
		
		P = new Vector* [usize];
		W = new float* [usize];
		
		for(int i=0; i<usize; i++) {
			P[i] = new Vector [vsize];
			W[i] = new float [vsize];
		}
		Tu = new float [uknots];
		Tv = new float [vknots];
		
		float f1, f2, f3;
		bool someknots = 1;
		int count = 0;
		while(count < 4 && someknots) {
			letter1 = getc(fptr);
			switch(letter1) {
			case 10:
			break;
			case 'p':
				cout << "reading points..."<<endl;
				for(int v=0; v<vsize; v++) {
					for(int u=0; u<usize; u++) {
						fscanf(fptr,"%f %f %f", &f1, &f2, &f3);
						P[u][v].x = f1;
						P[u][v].y = f2;
						P[u][v].z = f3;
						//cout << "P["<<u<<"]["<<v<<"] = ";
						//P[u][v].output();
					}
				}
				count++;
			break;
			case 'w':
				cout << "reading weights..." << endl;
				for(int v=0; v<vsize; v++) {
					for(int u=0; u<usize; u++) {
						fscanf(fptr,"%f", &f1);
						W[u][v] = f1;
						//cout << "W["<<u<<"]["<<v<<"] = " << f1 << endl;
					}
				}
				count++;
			break;
			case 'u':
				cout << "reading u knots..." << endl;
				for(int i=0; i<uknots; i++) {
					fscanf(fptr,"%f",&f1);
					Tu[i] = f1;
					//cout << "Tu["<<i<<"] = "<<Tu[i]<<endl;
				}
				count++;
			break;
			case 'v':
				cout << "reading v knots..." << endl;
				for(int i=0; i<uknots; i++) {
					fscanf(fptr,"%f",&f1);
					Tv[i] = f1;
					//cout << "Tv["<<i<<"] = "<<Tv[i]<<endl;
				}
				count++;
			break;
			case -1:
				cout << "Assuming uniform knot spacing" << endl;
				for(int i=0; i<udegree; i++) {
					Tu[i] = 0.0;
					Tu[uknots-i-1] = 1.0;
				}
				for(int i=udegree; i<uknots-udegree; i++) {
					Tu[i] = double(i-udegree)/double(uknots-2*udegree-1);
				}
				for(int i=0; i<vdegree; i++) {
					Tv[i] = 0.0;
					Tv[vknots-i-1] = 1.0;
				}
				for(int i=vdegree; i<vknots-vdegree; i++) {
					Tv[i] = double(i-vdegree)/double(vknots-2*vdegree-1);
				}
				someknots = 0;
			break;
			default:
				cout << "unknown character, "<<0+letter1<<"." << endl;
			}
		}
		fclose(fptr);
	}
/**************************************/
/*****END ROUTINE TO IMPORT FILE*******/
/**************************************/


/**************************************/
/****BEGIN ROUTINE TO EXPORT MESH******/
/**************************************/
	
	void output_mesh() {
		char outname [64];
		strcpy(outname,filehandle);
		strcat(outname,".inc");
		cout << "Outputting mesh to " << outname << endl;
		
		ofstream povout (outname);
		float u,v;
		Vector temp;
		povout << "vertex_vectors{ "<<ures*vres;
		for(int i=0; i<ures; i++) {
			u = double(i)/double(ures-1);
			for(int j=0; j<vres; j++) {
				v = double(j)/double(vres-1);
				temp = get_nurbs_point(u,v);
				povout << ",\n<"<<temp.x<<","<<temp.y<<","<<temp.z<<">";
			}
		}
		povout << "\n}\n";
		
		float delta = 0.001;
		Vector v1, v2, v3, v4;
		povout << "normal_vectors{ "<<ures*vres;
		for(int i=0; i<ures; i++) {
			u = double(i)/double(ures-1);
			for(int j=0; j<vres; j++) {
				v = double(j)/double(vres-1);
				v1 = get_nurbs_point(min(u+delta,Tu[uknots-1]),v);
				v2 = get_nurbs_point(max(u-delta,Tu[0]),v);
				v3 = get_nurbs_point(u,min(v+delta,Tv[vknots-1]));
				v4 = get_nurbs_point(u,max(v-delta,Tv[0]));
				temp = vnormalize(vcross(v1-v2,v3-v4));
				povout << ",\n<"<<temp.x<<","<<temp.y<<","<<temp.z<<">";
			}
		}
		povout << "\n}\n";
		
		
		povout << "uv_vectors{ "<<ures*vres;
		for(int i=0; i<ures; i++) {
			u = double(i)/double(ures-1);
			for(int j=0; j<vres; j++) {
				v = double(j)/double(vres-1);
				povout << ",\n<"<<u<<","<<v<<">";
			}
		}
		povout << "\n}\n";
		
		povout << "face_indices{ "<<(ures-1)*(vres-1)*2;
		for(int u=0; u<ures-1; u++) {
			for(int v=0; v<vres-1; v++) {
				povout << ",\n<"<<u*vres+v<<","<<(u+1)*vres+v<<","<<u*vres+v+1<<">";
				povout << ",\n<"<<(u+1)*vres+v<<","<<u*vres+v+1<<","<<(u+1)*vres+v+1<<">";
			}
		}
		povout << "\n}\n";
		
		povout << "normal_indices{ "<<(ures-1)*(vres-1)*2;
		for(int u=0; u<ures-1; u++) {
			for(int v=0; v<vres-1; v++) {
				povout << ",\n<"<<u*vres+v<<","<<(u+1)*vres+v<<","<<u*vres+v+1<<">";
				povout << ",\n<"<<(u+1)*vres+v<<","<<u*vres+v+1<<","<<(u+1)*vres+v+1<<">";
			}
		}
		povout << "\n}\n";
		
		povout << "uv_indices{ "<<(ures-1)*(vres-1)*2;
		for(int u=0; u<ures-1; u++) {
			for(int v=0; v<vres-1; v++) {
				povout << ",\n<"<<u*vres+v<<","<<(u+1)*vres+v<<","<<u*vres+v+1<<">";
				povout << ",\n<"<<(u+1)*vres+v<<","<<u*vres+v+1<<","<<(u+1)*vres+v+1<<">";
			}
		}
		povout << "\n}\n";
		
		
		povout.close();
	}
/**************************************/
/*****END ROUTINE TO EXPORT MESH*******/
/**************************************/


/**************************************/
/*******BEGIN U-NURBS FUNCTION*********/
/**************************************/
//first, find the point along one isoparm, then do that a few times and go the other way,
//if that makes any sense.  Something like interpolating as shown below.  Use De Boor's
//algorithm to insert a knot a number of times equal to the degree.  The last point
//is equal to the point on the curve.

//-----a-----
//     |
//-----a-----
//     b
//-----a-----
//     |
//-----a-----

	Vector4 get_u_nurbs_point(float u, int row) //De Boor's algorithm
	{
		int i=0;
		int j=usize;
		int k;
		bool again=1;
		while(again) { //search through an ordered list of knots
			k = (i+j)/2;
			if(u<Tu[k]) j=k;
			else i=k;
			if(i+1 >= j) again = 0;
		}
		k=i;
		Vector tP[udegree+1][udegree+1]; //set up hierarchy for inserting knots and finding points
		float tA[udegree+1][udegree+1];
		float tW[udegree+1][udegree+1];
		for(int c=0; c<=udegree; c++) {
			tP[c][0] = P[k-udegree+c][row]*W[k-udegree+c][row];
			tW[c][0] = W[k-udegree+c][row];
		}
		for(int r=1; r<udegree+1; r++) {
			for(int i=0; i<=udegree-r; i++) {
				tA[i][r] = (u-Tu[i+k-udegree+r])/(Tu[i+k+1]-Tu[i+k-udegree+r]);
				tP[i][r] = (1.0-tA[i][r])*tP[i][r-1]+tA[i][r]*tP[i+1][r-1];
				tW[i][r] = (1.0-tA[i][r])*tW[i][r-1]+tA[i][r]*tW[i+1][r-1];
			}
		}
		Vector4 out;
		out.x = tP[0][udegree].x/tW[0][udegree];
		out.y = tP[0][udegree].y/tW[0][udegree];
		out.z = tP[0][udegree].z/tW[0][udegree];
		out.t = tW[0][udegree];
		return out;
	}
/**************************************/
/*********END U-NURBS FUNCTION*********/
/**************************************/



/**************************************/
/*********BEGIN NURBS FUNCTION*********/
/**************************************/
	Vector get_nurbs_point(float u, float v) {
		int i=0;
		int j=vsize;
		int k;
		bool again=1;
		while(again) {
			k=(i+j)/2;
			if(v<Tv[k]) j=k;
			else i=k;
			if (i+1 >= j) again=0;
		}
		k=i;
		Vector tP[vdegree+1][vdegree+1];
		float tA[vdegree+1][vdegree+1];
		float tW[vdegree+1][vdegree+1];
		Vector4 temp;
		for(int c=0; c<=vdegree; c++) {
			temp = get_u_nurbs_point(u,k-vdegree+c);
			tP[c][0].x = temp.x*temp.t;
			tP[c][0].y = temp.y*temp.t;
			tP[c][0].z = temp.z*temp.t;
			tW[c][0] = temp.t;
		}
		for(int r=1; r<vdegree+1; r++) {
			for(int i=0; i<=vdegree-r; i++) {
				tA[i][r] = (v-Tv[i+k-vdegree+r])/(Tv[i+k+1]-Tv[i+k-vdegree+r]);
				tP[i][r] = (1.0-tA[i][r])*tP[i][r-1]+tA[i][r]*tP[i+1][r-1];
				tW[i][r] = (1.0-tA[i][r])*tW[i][r-1]+tA[i][r]*tW[i+1][r-1];
			}
		}
		Vector temp2 = tP[0][vdegree]/tW[0][vdegree];
		return temp2;
	}
};
/**************************************/
/***********END NURBS FUNCTION*********/
/**************************************/



/**************************************/
/************MAIN FUNCTION*************/
/**************************************/
//All we have to do is get a filename, import that into a surface, and export the mesh.
//Simple.

int main() {
	nurbs_surface S1;
	
	char fname [64];
	cout << "file to import: ";
	cin >> fname;
	S1.import_data(fname);
	//S1.import_data("points.txt");
	S1.output_mesh();
	
	return 0;
	
}
